package fg;
import nasm.*;
import util.graph.*;

import java.sql.SQLOutput;
import java.util.*;
import java.io.*;

public class Fg implements NasmVisitor <Void> {
    public Nasm nasm;
    public Graph graph;
    Map< NasmInst, Node> inst2Node;
    Map< Node, NasmInst> node2Inst;
    Map< String, NasmInst> label2Inst;

    public Fg(Nasm nasm){
	this.nasm = nasm;
	this.inst2Node = new HashMap< NasmInst, Node>();
	this.node2Inst = new HashMap< Node, NasmInst>();
	this.label2Inst = new HashMap< String, NasmInst>();
	this.graph = new Graph();

	/* on construit les noeuds */
        /**
         * parcourir chaque instruction dans la section de texte NASM.
         * Pour chaque instruction, il crée un nœud dans le graphe et ajoute la correspondance entre l'instruction et le nœud dans deux tables de hachage, inst2Node et node2Inst.
         * Il vérifie également si l'instruction a une étiquette et ajoute la correspondance entre l'étiquette et l'instruction dans une autre table de hachage, label2Inst.
         */
	for(NasmInst nasmInst : nasm.sectionText){
	    //	    System.out.println("<" + nasmInst.getClass().getSimpleName() + ">");
	    if(nasmInst.label != null) this.label2Inst.put(((NasmLabel)nasmInst.label).val, nasmInst);
	    Node n = this.graph.newNode();
	    this.inst2Node.put(nasmInst, n);
	    this.node2Inst.put(n, nasmInst);
	}
    /*
    parcourt chaque instruction dans la section de texte NASM, construit les arcs entre les nœuds du graphe.
    Il récupère le nœud correspondant à l'instruction en utilisant la table de hachage inst2Node, puis vérifie si l'instruction est un saut (jump) ou non.
    S'il s'agit d'un saut, il récupère l'étiquette de destination à partir de l'adresse de l'instruction et utilise la table de hachage label2Inst pour trouver l'instruction correspondante.
    Il récupère ensuite le nœud correspondant à l'instruction de destination en utilisant la table de hachage inst2Node, puis ajoute un arc du nœud courant au nœud de destination en utilisant la méthode graph.addEdge().
    Si l'instruction n'est pas un saut, l'algorithme ajoute un arc du nœud courant au nœud suivant dans la liste d'instructions en utilisant à nouveau la table de hachage inst2Node pour trouver le nœud correspondant.
    * */
	/* on construit les arcs */
	int instNb = nasm.sectionText.size();
        for (int i = 0; i < instNb; i++) {
            NasmInst nasmInst = nasm.sectionText.get(i);
            Node node = inst2Node.get(nasmInst);
            if(nasmInst instanceof NasmJne || nasmInst instanceof NasmJle || nasmInst instanceof NasmJge || nasmInst instanceof NasmJe || nasmInst instanceof NasmJmp || nasmInst instanceof NasmJl || nasmInst instanceof NasmJg ){
                String label = ((NasmLabel) nasmInst.address).val;
                Node dest = inst2Node.get(label2Inst.get(label));
                graph.addEdge(node , dest);
            }
            if(nasmInst instanceof NasmJmp || nasmInst instanceof NasmInt || nasmInst instanceof NasmRet ) continue;
            Node suivant = inst2Node.get(nasm.sectionText.get(i+1));
            graph.addEdge(node , suivant);

        }

	}



    public void affiche(String baseFileName){
	String fileName;
	PrintStream out = System.out;

	if (baseFileName != null){
	    try {
		baseFileName = baseFileName;
		fileName = baseFileName + ".fg";
		out = new PrintStream(fileName);
	    }
	    
	    catch (IOException e) {
		System.err.println("Error: " + e.getMessage());
	    }
	}
	
	for(NasmInst nasmInst : nasm.sectionText){
	    Node n = this.inst2Node.get(nasmInst);
	    out.print(n + " : ( ");
	    for(NodeList q=n.succ(); q!=null; q=q.tail) {
		out.print(q.head.toString());
		out.print(" ");
	    }
	    out.println(")\t" + nasmInst);
	}
	
    }




    public Void visit(NasmAdd inst){
       // if(inst.label != null) this.label2Inst.put(((NasmLabel)inst.label).val, inst);
       // Node myNode = graph.newNode();
       // inst2Node.put(inst, myNode);
       // node2Inst.put(myNode, inst);
       // NasmInst nasmInst = nasm.sectionText.get(i);
       // Node node = inst2Node.get(nasmInst);
       // Node sui = inst2Node.get(i+1);
       // graph.addEdge(node, sui);
        return null;
    }




    public Void visit(NasmMul inst){
        //if(inst.label != null) this.label2Inst.put(((NasmLabel)inst.label).val, inst);
        //Node myNode = this.graph.newNode();
        //inst2Node.put(inst, myNode);
        //node2Inst.put(myNode, inst);
        return null;
    }
    public Void visit(NasmSub inst){
        //if(inst.label != null) this.label2Inst.put(((NasmLabel)inst.label).val, inst);
        //Node myNode = graph.newNode();
        //inst2Node.put(inst, myNode);
        //node2Inst.put(myNode, inst);
        return null;
    }
    public Void visit(NasmCall inst){
       // if(inst.label != null) this.label2Inst.put(((NasmLabel)inst.label).val, inst);
       // Node myNode = graph.newNode();
       // inst2Node.put(inst, myNode);
       // node2Inst.put(myNode, inst);
        return null;
    }
    public Void visit(NasmDiv inst){
        /*
        if(inst.label != null) this.label2Inst.put(((NasmLabel)inst.label).val, inst);
        Node myNode = graph.newNode();
        inst2Node.put(inst, myNode);
        node2Inst.put(myNode, inst);*/
        return null;
    }
    public Void visit(NasmJe inst){
        /*
        Node node = inst2Node.get(inst);
        String label = ((NasmLabel) inst.address).val;
        Node to = inst2Node.get(label2Inst.get(label));
        graph.addEdge(node,to);*/
        return null;
    }
    public Void visit(NasmJle inst){
        /*
        Node node = inst2Node.get(inst);
        String label = ((NasmLabel) inst.address).val;
        Node to = inst2Node.get(label2Inst.get(label));
        graph.addEdge(node,to);*/
        return null;
    }
    public Void visit(NasmJne inst){
       /* Node node = inst2Node.get(inst);
        String label = ((NasmLabel) inst.address).val;
        Node to = inst2Node.get(label2Inst.get(label));
        graph.addEdge(node,to);*/
        return null;}

    public Void visit(NasmOr inst){return null;}
    public Void visit(NasmCmp inst){return null;}
    public Void visit(NasmInst inst){return null;}
    public Void visit(NasmJge inst){
       /* Node node = inst2Node.get(inst);
        String label = ((NasmLabel) inst.address).val;
        Node to = inst2Node.get(label2Inst.get(label));
        graph.addEdge(node,to);*/
        return null;
    }
    public Void visit(NasmJl inst){return null;}
    public Void visit(NasmNot inst){return null;}
    public Void visit(NasmPop inst){return null;}
    public Void visit(NasmRet inst){return null;}
    public Void visit(NasmXor inst){return null;}
    public Void visit(NasmAnd inst){return null;}
    public Void visit(NasmJg inst){return null;}
    public Void visit(NasmJmp inst){return null;}
    public Void visit(NasmMov inst){
       /* if(inst.label != null) this.label2Inst.put(((NasmLabel)inst.label).val, inst);
        Node myNode = graph.newNode();
        inst2Node.put(inst, myNode);
        node2Inst.put(myNode, inst);*/
        return null;
    }
    public Void visit(NasmPush inst){return null;}
    public Void visit(NasmEmpty inst){return null;}
    public Void visit(NasmInt inst){return null;}

    public Void visit(NasmAddress operand){return null;}
    public Void visit(NasmConstant operand){return null;}
    public Void visit(NasmLabel operand){return null;}
    public Void visit(NasmRegister operand){return null;}

    public Void visit(NasmResb pseudoInst){return null;}
    public Void visit(NasmResw pseudoInst){return null;}
    public Void visit(NasmResd pseudoInst){return null;}
    public Void visit(NasmResq pseudoInst){return null;}
    public Void visit(NasmRest pseudoInst){return null;}

}
