package nasm;
import c3a.*;
import ts.*;


public class C3a2nasm implements C3aVisitor <NasmOperand> {
	private C3a c3a;
	private Nasm nasm;
	private Ts tableGlobale;
	private TsItemFct currentFct;
	private NasmRegister esp;
	private NasmRegister ebp;


	public C3a2nasm(C3a c3a, Ts tableGlobale){
		this.c3a = c3a;
		nasm = new Nasm(tableGlobale);
		nasm.setTempCounter(c3a.getTempCounter());

		this.tableGlobale = tableGlobale;
		this.currentFct = null;
		esp = new NasmRegister(-1);
		esp.colorRegister(Nasm.REG_ESP);

		ebp = new NasmRegister(-1);
		ebp.colorRegister(Nasm.REG_EBP);

		NasmOperand res;
		for(C3aInst c3aInst : c3a.listeInst){
			//	   	    System.out.println("<" + c3aInst.getClass().getSimpleName() + ">");
			res = c3aInst.accept(this);
		}
	}

	public Nasm getNasm(){return nasm;}

    /*--------------------------------------------------------------------------------------------------------------
      transforme une opérande trois adresses en une opérande asm selon les règles suivantes :

      C3aConstant -> NasmConstant
      C3aTemp     -> NasmRegister
      C3aLabel    -> NasmLabel
      C3aFunction -> NasmLabel
      C3aVar      -> NasmAddress
      --------------------------------------------------------------------------------------------------------------*/

	@Override
	public NasmOperand visit(C3aTemp temp){

		return new NasmRegister(temp.num);
	}

	@Override
	public NasmOperand visit(C3aVar oper) {
		if(oper.item.portee == tableGlobale){
			if(oper.index == null) return new NasmAddress(new NasmLabel(oper.item.identif));
			NasmOperand register = nasm.newRegister();
			nasm.ajouteInst(new NasmMov(null,register, oper.index.accept(this) , ""));
			nasm.ajouteInst(new NasmMul(null , register , new NasmConstant(4), ""));
			return new NasmAddress(new NasmLabel(oper.item.identif) , '+' , register);
		}
		char dest = oper.item.isParam ? '+' : '-';
		int register = oper.item.isParam ? 8 + oper.item.portee.getAdrArgCourante() - oper.item.getAdresse() : 4 + oper.item.getAdresse();
		return new NasmAddress(ebp , dest , new NasmConstant(register));
		/*if (oper.index == null) {
			if (oper.item.isParam) {
				return new NasmAddress(ebp, '+', new NasmConstant(8 + (oper.item.portee.getAdrArgCourante()) - oper.item.adresse));
			} else if (!oper.item.portee.equals(tableGlobale)) {
				return new NasmAddress(ebp, '-', new NasmConstant(4 + oper.item.adresse));
			} else {

				return new NasmAddress(new NasmLabel(oper.item.identif));
			}
		}
		NasmOperand index = oper.index.accept(this);
		NasmRegister register = nasm.newRegister();
		nasm.ajouteInst(new NasmMov(null,register,index,""));
		nasm.ajouteInst(new NasmMul(null,register,new NasmConstant(4),""));
		return new NasmAddress(new NasmLabel(oper.item.identif),'+',register);
		*/

	}

	@Override
	public NasmOperand visit(C3aFunction oper) {

		return new NasmLabel(oper.toString());
	}

	/*--------------------------------------------------------------------------------------------------------------*/
	@Override
	public NasmOperand visit(C3aInstAdd inst)
	{
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmOperand oper1 = inst.op1.accept(this);
		NasmOperand oper2 = inst.op2.accept(this);
		NasmOperand dest = inst.result.accept(this);
		nasm.ajouteInst(new NasmMov(label, dest, oper1, ""));
		nasm.ajouteInst(new NasmAdd(null , dest, oper2, ""));

		return null;

	}

	@Override
	public NasmOperand visit(C3aInstCall inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		//NasmOperand op1 = inst.op1.accept(this);
		//NasmOperand result = inst.result.accept(this);

		nasm.ajouteInst(new NasmSub(label , esp , new NasmConstant(4) , "Alloc de mémoir de la valeur de retour" ));
		nasm.ajouteInst(new NasmCall(null , inst.op1.accept(this) , ""));

		nasm.ajouteInst(new NasmPop(null, inst.result.accept(this) , "recupére la valeur de retour"));
		if(inst.op1.val.getNbArgs()!=0) {
			nasm.ajouteInst(new NasmAdd(null, esp, new NasmConstant(inst.op1.val.table.getAdrArgCourante()), "#libération"));
		}


		return null;
	}

	@Override
	public NasmOperand visit(C3aInstFBegin inst) {
		currentFct = inst.val;
		NasmOperand label = new NasmLabel(inst.val.identif);

		nasm.ajouteInst(new NasmPush(label,ebp,"sauvegarde la valeur de ebp"));
		nasm.ajouteInst(new NasmMov(null,ebp,esp,"nouvelle valeur de ebp"));
		NasmRegister reg_eax = nasm.newRegister();
		reg_eax.colorRegister(Nasm.REG_EAX);

		NasmRegister reg_ebx = nasm.newRegister();
		reg_ebx.colorRegister(Nasm.REG_EBX);

		NasmRegister reg_ecx = nasm.newRegister();
		reg_ecx.colorRegister(Nasm.REG_ECX);

		NasmRegister reg_edx = nasm.newRegister();
		reg_edx.colorRegister(Nasm.REG_EDX);



		nasm.ajouteInst(new NasmPush(null,reg_eax,"sauvegarde de eax"));

		nasm.ajouteInst(new NasmPush(null,reg_ebx,"sauvegarde de ebx"));
		nasm.ajouteInst(new NasmPush(null,reg_ecx,"sauvegarde de ecx"));
		nasm.ajouteInst(new NasmPush(null,reg_edx,"sauvegarde de edx"));


		int nbVar = inst.val.saDecFonc.getVariable() == null?0:inst.val.saDecFonc.getVariable().length();
		nasm.ajouteInst(new NasmSub(null,esp,new NasmConstant(4*nbVar),"allocation des variables locales"));
		return null;
		//currentFct = inst.val;
		//NasmOperand label = new NasmLabel(currentFct.getIdentif()) ;
		//NasmRegister reg_ebp = new NasmRegister(Nasm.REG_EBP);
		//reg_ebp.colorRegister(Nasm.REG_EBP);
		//NasmRegister reg_esp = new NasmRegister(Nasm.REG_ESP);
		//reg_esp.colorRegister(Nasm.REG_ESP);
		//nasm.ajouteInst(new NasmPush(label , ebp , ""));
		//nasm.ajouteInst(new NasmMov(null,ebp,esp,""));
		//nasm.ajouteInst(new NasmSub(null , esp , new  NasmConstant(currentFct.table.getAdrVarCourante()) , "" ) );
	}

	@Override
	public NasmOperand visit(C3aInst inst) {
		inst.accept(this);
		return null;
	}

	@Override
	public NasmOperand visit(C3aInstJumpIfLess inst)
	{
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		//NasmOperand operand1 = inst.op1.accept(this);
		//NasmOperand operand2 = inst.op2.accept(this);
		//NasmOperand result = inst.result.accept(this);

		//nasm.ajouteInst(new NasmMov(label , result , operand1 , ""));
		nasm.ajouteInst(new NasmCmp( null , inst.op1.accept(this) , inst.op2.accept(this) , "" ));
		nasm.ajouteInst(new NasmJl(null , inst.result.accept(this) , "" ));
		return null;
	}

	@Override
	public NasmOperand visit(C3aInstMult inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmOperand oper1 = inst.op1.accept(this);
		NasmOperand oper2 = inst.op2.accept(this);
		NasmOperand dest = inst.result.accept(this);
		nasm.ajouteInst(new NasmMov(label, dest, oper1, ""));
		nasm.ajouteInst(new NasmMul(null , dest, oper2, ""));
		return null;
	}

	@Override
	public NasmOperand visit(C3aInstRead inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmRegister reg_eax = nasm.newRegister();
		reg_eax.colorRegister(Nasm.REG_EAX);
		//NasmOperand op = inst.result.accept(this);
		nasm.ajouteInst(new NasmMov(label , reg_eax , new NasmLabel("sinput") , ""));
		nasm.ajouteInst(new NasmCall(null , new NasmLabel("readline") , "readline"));
		nasm.ajouteInst(new NasmCall(null , new NasmLabel("atoi") , "atoi"));
		nasm.ajouteInst(new NasmMov(null , inst.result.accept(this) , reg_eax , "" ));

		return null;
	}

	@Override
	public NasmOperand visit(C3aInstSub inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmOperand oper1 = inst.op1.accept(this);
		NasmOperand oper2 = inst.op2.accept(this);
		NasmOperand dest = inst.result.accept(this);
		nasm.ajouteInst(new NasmMov(label, dest, oper1, ""));
		nasm.ajouteInst(new NasmSub(null , dest, oper2, ""));
		return null;
	}

	@Override
	public NasmOperand visit(C3aInstAffect inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		//NasmOperand oper1 = inst.op1.accept(this);
		//NasmOperand resultat = inst.result.accept(this);
		nasm.ajouteInst(new NasmMov(label , inst.result.accept(this) , inst.op1.accept(this) , "Affectation"));
		return null;
	}

	@Override
	public NasmOperand visit(C3aInstDiv inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		//NasmOperand op1 = inst.op1.accept(this);
		NasmOperand op2 = inst.op2.accept(this);
		//NasmRegister reg_eax = nasm.newRegister();
		//reg_eax.colorRegister(Nasm.REG_EAX);
		NasmRegister eax = nasm.newRegister();
		eax.colorRegister(Nasm.REG_EAX);
		NasmRegister edx = nasm.newRegister();
		edx.colorRegister(Nasm.REG_EDX);
		//nasm.ajouteInst(new NasmMov(label, reg_eax,op1, ""));
		nasm.ajouteInst(new NasmMov(label, eax,inst.op1.accept(this), ""));
		nasm.ajouteInst(new NasmMov(null, edx,new NasmConstant(0), ""));

		if (op2 instanceof NasmConstant){
			op2 = nasm.newRegister();
			nasm.ajouteInst(new NasmMov(null, op2, inst.op2.accept(this) , ""));
		}
		nasm.ajouteInst(new NasmDiv(null , op2 , ""));
		nasm.ajouteInst(new NasmMov(null , edx , edx , ""));
		nasm.ajouteInst(new NasmMov(null , inst.result.accept(this) , eax , "" ));
			//nasm.ajouteInst(new NasmDiv(null,inst.op2.accept(this),"div op2 par la valeur de reg_eax"));
		/*else {

			NasmRegister reg_ebx = nasm.newRegister();

			nasm.ajouteInst(new NasmMov(null, reg_ebx,op2, "mettre op2 dans reg_ebx"));
			nasm.ajouteInst(new NasmDiv(null,reg_ebx,"diviser par la valeur de reg_ebx"));

		}*:
		NasmRegister reg_eax1 = nasm.newRegister();
		reg_eax1.colorRegister(Nasm.REG_EAX);
		nasm.ajouteInst(new NasmMov(null,inst.result.accept(this),reg_eax1,""));
		/*NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmOperand oper1 = inst.op1.accept(this);
		NasmOperand oper2 = inst.op2.accept(this);
		NasmRegister reg_eax = nasm.newRegister();
		reg_eax.colorRegister(Nasm.REG_EAX);
		NasmRegister reg_edx = nasm.newRegister();
		reg_edx.colorRegister(Nasm.REG_EBX);
		// Stocke la valeur du premier opérande dans EAX
		nasm.ajouteInst(new NasmMov(label,reg_eax, oper1, ""));
		nasm.ajouteInst(new NasmMov(label,reg_edx, new NasmConstant(0), ""));
		nasm.ajouteInst(new NasmDiv(null, reg_edx, ""));
        */
		return null;
	}

	@Override
	public NasmOperand visit(C3aInstFEnd inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;


		NasmRegister reg_eax = nasm.newRegister();
		reg_eax.colorRegister(Nasm.REG_EAX);

		NasmRegister reg_ebx = nasm.newRegister();
		reg_ebx.colorRegister(Nasm.REG_EBX);

		NasmRegister reg_ecx = nasm.newRegister();
		reg_ecx.colorRegister(Nasm.REG_ECX);

		NasmRegister reg_edx = nasm.newRegister();
		reg_edx.colorRegister(Nasm.REG_EDX);

		int nbrVar = currentFct.saDecFonc.getVariable()!=null ? currentFct.saDecFonc.getVariable().length() : 0;
		nasm.ajouteInst(new NasmAdd(label,esp,new NasmConstant(nbrVar*4),"désallocation des variables locales"));


		nasm.ajouteInst(new NasmPop(null,reg_edx,"restaure edx"));
		nasm.ajouteInst(new NasmPop(null,reg_ecx,"restaure ecx"));
		nasm.ajouteInst(new NasmPop(null,reg_ebx,"restaure ebx"));
		nasm.ajouteInst(new NasmPop(null,reg_eax,"restaure eax"));


		nasm.ajouteInst(new NasmPop(null,ebp,"restaure la valeur de ebp"));
		nasm.ajouteInst(new NasmRet(null,""));
		return null;
		/*NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmRegister reg_esp = new NasmRegister(Nasm.REG_ESP);
		reg_esp.colorRegister(Nasm.REG_ESP);
		nasm.ajouteInst(new NasmAdd(label,reg_esp,new NasmConstant(4), ""));
		NasmRegister reg_ebp = new NasmRegister(Nasm.REG_EBP);
		reg_ebp.colorRegister(Nasm.REG_EBP);
		nasm.ajouteInst(new NasmPop(null , reg_ebp , ""));
		nasm.ajouteInst(new NasmRet(null , ""));*/
	}

	@Override
	public NasmOperand visit(C3aInstJumpIfEqual inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		//NasmRegister reg_eax = nasm.newRegister();
		nasm.ajouteInst(new NasmCmp(label,inst.op1.accept(this), inst.op2.accept(this), ""));
		//nasm.ajouteInst(new NasmCmp(null,reg_eax,inst.op2.accept(this),"passer par un registre temporaire"));
		nasm.ajouteInst(new NasmJe(null,inst.result.accept(this),""));
		return null;
	}

	@Override
	public NasmOperand visit(C3aInstJumpIfNotEqual inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmOperand oper1 = inst.op1.accept(this);
		NasmOperand oper2 = inst.op2.accept(this);
		NasmOperand resultat = inst.result.accept(this);
		NasmRegister reg_eax = nasm.newRegister();
		nasm.ajouteInst(new NasmMov(label,reg_eax,oper1,""));
		nasm.ajouteInst(new NasmCmp(null,oper1, oper2, ""));
		nasm.ajouteInst(new NasmJne(null,resultat,"jumpIfNotEq"));

		return null;
	}

	@Override
	public NasmOperand visit(C3aInstJump inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmOperand resultat = inst.result.accept(this);
		nasm.ajouteInst(new NasmJmp(label,resultat,"Jump"));
		return null;
	}

	@Override
	public NasmOperand visit(C3aInstParam inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmOperand op1 = inst.op1.accept(this);
		nasm.ajouteInst(new NasmPush(label,op1,"Param"));
		return null;
	}

	@Override
	public NasmOperand visit(C3aInstReturn inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmOperand op = inst.op1.accept(this);
		//NasmRegister reg_ebp = new NasmRegister(Nasm.REG_EBP+8);
		//reg_ebp.colorRegister(Nasm.REG_EBP);
		nasm.ajouteInst(new NasmMov(label , new NasmAddress(ebp, '+' , new NasmConstant(8)) , op , "" ));
		//nasm.ajouteInst(new NasmRet(label , ""));

		return null;
	}

	@Override
	public NasmOperand visit(C3aInstWrite inst) {

		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		//NasmOperand resultat = inst.op1.accept(this);
		NasmRegister reg_eax = nasm.newRegister();
		reg_eax.colorRegister(Nasm.REG_EAX);
		nasm.ajouteInst(new NasmMov(label , reg_eax , inst.op1.accept(this) , "write 1"));
		nasm.ajouteInst(new NasmCall(null , new NasmLabel("iprintLF") , "write 2"));

		return null;
	}

	@Override
	public NasmOperand visit(C3aInstStop inst) {
		NasmOperand label = (inst.label != null) ? inst.label.accept(this) : null;
		NasmRegister reg_eax = nasm.newRegister();
		reg_eax.colorRegister(Nasm.REG_EAX);
		NasmRegister reg_ebx = nasm.newRegister();
		reg_ebx.colorRegister(Nasm.REG_EBX);
		nasm.ajouteInst(new NasmMov(label,reg_ebx,new NasmConstant(0),""));
		nasm.ajouteInst(new NasmMov(label,reg_eax,new NasmConstant(1),""));
		nasm.ajouteInst(new NasmInt(null , ""));
		return null;
	}

	@Override
	public NasmOperand visit(C3aConstant oper)
	{
		return new NasmConstant((oper.val));
	}

	@Override
	public NasmOperand visit(C3aBooleanConstant oper) {

		//return new NasmConstant(oper.val ? "1" : "0");
		return  new NasmConstant(oper.val ? 1 : 0) ; //new NasmConstant(0);
	}

	@Override
	public NasmOperand visit(C3aLabel oper) {
		return new NasmLabel(oper.toString());
	}


}
