package sa;
import ts.*;
import util.Error;
import util.Type;

public class SaTypeCheck extends SaDepthFirstVisitor <Void>{
    private TsItemFct fonctionCourante;

    public SaTypeCheck(SaNode root)
    {
	try{
	root.accept(this);
	}
	catch(ErrorException e){
	    System.err.print("ERREUR DE TYPAGE : ");
	    System.err.println(e.getMessage());
	    System.exit(e.getCode());
	}
	catch(Exception e){}
    }

    @Override
    public void defaultIn(SaNode node)
    {
	//			System.out.println("<" + node.getClass().getSimpleName() + ">");
    }

    @Override
    public void defaultOut(SaNode node)
    {
	//		System.out.println("</" + node.getClass().getSimpleName() + ">");
    }
    /////////////////////////////////////////////////////
    @Override
    public Void visit(SaExpVrai node) throws Exception
    {
        defaultIn(node);
        Type type = node.getType();
        if(type != Type.BOOL)
            throw new ErrorException(Error.TYPE ,"Vérifier bien si le type est bien de type bool ");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpFaux node) throws Exception
    {
        defaultIn(node);
        Type type = node.getType();
        if(type != Type.BOOL)
            throw new ErrorException(Error.TYPE , "Vérifier bien si le type est de type bool");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpInt node) throws Exception
    {
        defaultIn(node);
        Type type = node.getType();
        if(type!=Type.ENTIER)
            throw new ErrorException(Error.TYPE , "Vérifier bien si le type est de type entier");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpMult node) throws Exception
    {
        defaultIn(node);
        Type type = node.getType();
        Type type1 = node.getOp1().getType();
        Type type2 = node.getOp2().getType();
        if(!Type.checkCompatibility(type,type1) || !Type.checkCompatibility(type,type2))
            throw new ErrorException(Error.TYPE , "Erreur de compatibilty des Opérandes");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpDiv node) throws Exception
    {
        defaultIn(node);
        Type type = node.getType();
        Type type1 = node.getOp1().getType();
        Type type2 = node.getOp2().getType();
        if(!Type.checkCompatibility(type,type1) || !Type.checkCompatibility(type,type2))
            throw new ErrorException(Error.TYPE , "Erreur de type Opérande");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaDecFonc node) throws Exception
    {
        defaultIn(node);
        this.fonctionCourante = node.tsItem;
        if(node.getParametres() != null) node.getParametres().accept(this);
        if(node.getVariable() != null) node.getVariable().accept(this);
        if(node.getCorps() != null) node.getCorps().accept(this);
        this.fonctionCourante = null;
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpAdd node) throws Exception
    {
        defaultIn(node);
        Type t = node.getType();
        Type t1 = node.getOp1().getType();
        Type t2 = node.getOp2().getType();
        if(!Type.checkCompatibility(t,t1) || !Type.checkCompatibility(t,t2))
            throw new ErrorException(Error.TYPE , "Erreur de compatibilite des opérandes d'addition ");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpSub node) throws Exception
    {
        defaultIn(node);
        Type t = node.getType();
        Type t1 = node.getOp1().getType();
        Type t2 = node.getOp2().getType();
        if(!Type.checkCompatibility(t,t1) || !Type.checkCompatibility(t,t2))
            throw new ErrorException(Error.TYPE , "Erreur de compatibilite des opérandes soustraction ");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpOr node) throws Exception
    {
        defaultIn(node);
        Type type = node.getType();
        Type type1 = node.getOp1().getType();
        Type type2 = node.getOp2().getType();
        if(!Type.checkCompatibility(type ,type1) || !Type.checkCompatibility(type,type2))
            throw new ErrorException(Error.TYPE , "Vérifier bien si  les opérandes sont compatibles");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpNot node) throws Exception
    {
        defaultIn(node);
        Type type = node.getType();
        Type type1 = node.getOp1().getType();
        if(!Type.checkCompatibility(type,type1))
            throw new ErrorException(Error.TYPE , "Vérifier que les opérandes sont bien compatible");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpLire node) throws Exception
    {
        defaultIn(node);
        Type type = node.getType();
        if(type != Type.ENTIER)
            throw new ErrorException(Error.TYPE ,"Erreur de Type");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaInstSi node) throws Exception //jump
    {
        defaultIn(node);
        Type test = node.getTest().getType();
        if(!Type.checkCompatibility(test,Type.BOOL))
            throw new ErrorException(Error.TYPE , "incompatibility  du type de test");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaInstAffect node) throws Exception
    {
        defaultIn(node);
        Type t1 = node.getLhs().getTsItem().getType();
        Type t2 = node.getRhs().getType();
        if(!Type.checkCompatibility(t1,t2))
            throw new ErrorException(Error.TYPE, "erreur de compatibilté d'affectation");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaInstTantQue node) throws Exception
    {
        defaultIn(node);
        Type t =  node.getTest().getType();
        if(!Type.checkCompatibility(t,Type.BOOL))
            throw new ErrorException(Error.TYPE , "Erreur de compatibilité du test TQ");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaInstRetour node) throws Exception
    {
        defaultIn(node);
        if(!Type.checkCompatibility(node.getVal().getType() ,this.fonctionCourante.typeRetour))
            throw new ErrorException(Error.TYPE , "Erreur de Compatibilité de type de retour");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpAnd node) throws Exception
    {
        defaultIn(node);
        Type type = node.getType();
        Type type1 = node.getOp1().getType();
        Type type2 = node.getOp2().getType();
        if(!Type.checkCompatibility(type ,type1) || !Type.checkCompatibility(type,type2))
            throw new ErrorException(Error.TYPE , "Vérifier bien si  les opérandes sont compatibles");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpInf node) throws Exception
    {
        defaultIn(node);
        Type type1 = node.getOp1().getType();
        Type type2 = node.getOp2().getType();
        if(!Type.checkCompatibility(type1 ,Type.ENTIER) || !Type.checkCompatibility(type2,Type.ENTIER))
            throw new ErrorException(Error.TYPE , "Vérifier bien si  les opérandes sont comparable");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpEqual node) throws Exception
    {
        defaultIn(node);
        Type type1 = node.getOp1().getType();
        Type type2 = node.getOp2().getType();
        if(!Type.checkCompatibility(type1,type2))
            throw new ErrorException(Error.TYPE , "Impossible de faire l'égalité de deux type différent");
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaAppel node) throws Exception
    {
        defaultIn(node);

        SaLExp arguments = node.getArguments();
        SaLDecVar decParam = node.tsItem.saDecFonc.getParametres();
        while(arguments != null) {
            //Type type1= arguments.getTete().getType();
            //Type type2 = decParam.getTete().getType();
            if(!Type.checkCompatibility(arguments.getTete().getType(),decParam.getTete().getType()))
                throw new ErrorException(Error.TYPE , "Erreur de Type Arguments");
            arguments = arguments.getQueue();
            decParam = decParam.getQueue();

        }
        defaultOut(node);
        return null;
    }
    @Override
    public Void visit(SaExpAppel node) throws Exception
    {
	defaultIn(node);
	node.getVal().accept(this);
	defaultOut(node);
	return null;
    }
    @Override
    public Void visit(SaVarIndicee node) throws Exception
    {
        defaultIn(node);
        if(!Type.checkCompatibility(node.getIndice().getType() , Type.ENTIER))
            throw  new ErrorException(Error.TYPE ,"Vérifier que l'indice est bien de type entier");
        defaultOut(node);
        return null;
    }




}
