package ts;
import sa.*;
import util.Error;
import util.Type;

public class Sa2ts extends SaDepthFirstVisitor <Void> {
    enum Context {
	LOCAL,
	GLOBAL,
	PARAM
    }
    
    private Ts tableGlobale;
    private Ts tableLocaleCourante;
    private Context context;
    
    public Ts getTableGlobale(){return this.tableGlobale;}
    public Sa2ts(SaNode root)
    {
	tableGlobale = new Ts();
	tableLocaleCourante = null;
	context = Context.GLOBAL;
	try{
	    root.accept(this);
	    if(tableGlobale.getFct("main") == null)
		throw new ErrorException(Error.TS, "la fonction main n'est pas définie");
	}
	catch(ErrorException e){
	    System.err.print("ERREUR TABLE DES SYMBOLES : ");
	    System.err.println(e.getMessage());
	    System.exit(e.getCode());
	}
	catch(Exception e){}
    }
	/////////////////////////////////////////////////////////////
	@Override
	public Void visit(SaDecFonc node) throws Exception
	{
		String nom = node.getNom();
		//Type typeRetour = node.getTypeRetour();
		tableLocaleCourante = new Ts();
		defaultIn(node);
		if (tableGlobale.getFct(nom)!=null)
			throw new ErrorException(Error.TS , "La fonction existe déjà dans la table global");
		int taille = node.getParametres() == null ? 0 : node.getParametres().length();
		node.tsItem = tableGlobale.addFct(nom,node.getTypeRetour(),taille,tableLocaleCourante,node);
		context = Context.PARAM;
		if(node.getParametres() != null) node.getParametres().accept(this);
		context = Context.LOCAL;
		if(node.getVariable() != null) node.getVariable().accept(this);
		if(node.getCorps() != null ) node.getCorps().accept(this);
		context = context.GLOBAL;
		tableLocaleCourante = null;
		//if(node.getCorps() != null) node.getCorps().accept(this);
		defaultOut(node);
		return null;

	}
	@Override
	public Void visit(SaVarSimple node) throws Exception
	{
		TsItemVar item = null;
		defaultIn(node);
		if(context == Context.LOCAL || context == Context.PARAM){
			item = tableLocaleCourante.getVar(node.getNom());
			if(item != null){
				node.tsItem = (TsItemVarSimple) item;
			}

			else{
				item = tableGlobale.getVar(node.getNom());
				if(item!=null){
					node.tsItem = (TsItemVarSimple) item;
				}
				else{
					throw new ErrorException(Error.TS , "La variable n'existe pas (elle est ni LOCAL ni GLOBAL)");
				}
			}
		}
		defaultOut(node);
		return null;
	}
	@Override
	public Void visit(SaDecVar node) throws Exception
	{
		defaultIn(node);
		TsItemVar item = null;
		if(context==Context.LOCAL){
			if(tableLocaleCourante.getVar(node.getNom())==null){
				item = tableLocaleCourante.addVar(node.getNom(),node.getType());
				if(item != null) node.setTsItem(item);
			}else{
				throw new ErrorException(Error.TS , "Variable existant en Local" );
			}
		}
		if(context == Context.GLOBAL){
			if(tableGlobale.getVar(node.getNom())==null){
				item = tableGlobale.addVar(node.getNom() , node.getType());
				if(item!=null){
					node.setTsItem(item);
				}
			}else {
				throw new ErrorException(Error.TS , "Variable existant en Local");
			}
		}
		if(context == Context.PARAM){
			if(tableLocaleCourante.getVar(node.getNom())==null){
				item = tableLocaleCourante.addParam(node.getNom() , node.getType());
				if(item != null){
					node.setTsItem(item);
				}
			}
			else {
				throw  new ErrorException(Error.TS , "Le paramètre existe déja");
			}
		}

		defaultOut(node);
		return null;
	}
	@Override
	public Void visit(SaDecTab node) throws Exception{
		defaultIn(node);
		if(context != Context.GLOBAL)
			throw new ErrorException(Error.TS , "Le tableau est local");
		if(tableGlobale.getVar(node.getNom())!=null)
			throw new ErrorException(Error.TS, "Tableau existant dans la table Global");

		node.setTsItem(tableGlobale.addTab(node.getNom(), node.getType(), node.getTaille()));
		defaultOut(node);
		return null;
	}


	@Override
	public Void visit(SaVarIndicee node) throws Exception
	{
		String nom = node.getNom();
		//SaExp indice = node.getIndice();
		defaultIn(node);
		if(tableGlobale.getVar(nom)!=null) {
			node.tsItem = tableGlobale.getVar(nom);
			node.getIndice().accept(this);
			defaultOut(node);
			return null;
		}
		else
			throw new ErrorException(Error.TS , "Tableau inexistant");
	}
	@Override
	public Void visit(SaAppel node) throws Exception
	{
		/*defaultIn(node);
		int taille = 0;
		node.tsItem = tableGlobale.getFct(node.getNom());
		if(node.tsItem==null)
			throw new ErrorException(Error.TS , "Fonction non déclarée");
		if(node.tsItem.getNbArgs()!= tableGlobale.getFct(node.getNom()).getNbArgs()){
			throw new ErrorException(Error.TS , "Nbre arg incorrect");
		}
		defaultOut(node);
		return null*/
		//String nom = node.getNom();
		defaultIn(node);
		if(tableGlobale.getFct(node.getNom())==null)
			throw new ErrorException(Error.TS , "Fonction non déclarée");

		if (node.getArguments() != null) {
			if(node.getArguments().length() != tableGlobale.getFct(node.getNom()).getNbArgs())
				throw new ErrorException(Error.TS , "Error Nbre argument");
			node.getArguments().accept(this);
		}
		else
		{
			if(tableGlobale.getFct(node.getNom()).getNbArgs()!=0)
				throw new ErrorException(Error.TS , "Error nbre Argument");
		}
		node.tsItem = tableGlobale.getFct(node.getNom());
		defaultOut(node);
		return null;
  // graziella lauryne
	}

	/////////////////////////////
	@Override
	public Void visit(SaInstSi node) throws Exception
	{
		defaultIn(node);
		node.getTest().accept(this);
		node.getAlors().accept(this);
		if(node.getSinon() != null) node.getSinon().accept(this);
		defaultOut(node);
		return null;
	}


}
